#ifndef VECTOR_HPP
#define VECTOR_HPP

#include <windows.h>   // must include this before gl.h
#include <gl/gl.h> //for  OpenGL methods like glVertex()
#include "realnumber.h"
#include <iostream.h> //for the istream and ostream classes.
#include <math.h> //for sqrt

/* 
History:

	The cVectorTransformation.* files implement 2 and 3 dimensional vectors and some
matrices for transforming them.  They were developed by Rudy Rucker for CS 116A at SJSU 
in Spring, 1996 in a series of 5 versions, up through cVector5.h and cVector5.cpp.
	This cVectorTransformation version was developed in Spring, 1999, for use in 
Rucker's SOFTWARE PROJECTS text.  It was extensively rewritten at this time.
	The code in this module was very heavily reworked again in Spring/Summer 2001
as part of the port to 3D OpenGL from our old 2D Windows API graphics.  These
graphics versions are called cGraphicsMFC and cGraphicsOpenGL; see the
graphics.h header for more information.
    In Fall, 2003, I took out the special MFC classes like CArchive and types like BOOL
and made it compatible with ANSI C. For use with OpenGL in CS 116a, I also dropped the
cVector2 and cMatrix2 classes, and added some OpenGL specific methods.

Usage:

	The cVector3 classes represent three dimensional vectors.
The +, - and * operators for addition, subtration, and scalar multiplication.
The operator% is used for the dot product.  Cross product has not been implemented.
	The cMatrix3 classes are used as transformations on,
cVector3 objects.  To apply the transformation to a 
vector we use v.apply(M) or the overloaded friend operator* as in v = M * u.
Always use parentheses when you have several transformations so as to enforce the
desired order of evaluation.
	Our transformations effectively perform a scaling, shear and/or a rotation
followed by a translation.  When you multiply two transformations to get 
S = R * T, the effect of S will be to perform T and to then perform R.

Implementation:
	We think of a 3D vector as a 1 by 4 column matrix with a 1.0 in the bottom
row, and we think of a cMatrix3 as a 4 by 4 matrix with a bottom 
row of 0.0  0.0  0.0 1.0.
	The reason we do this is so that we can fit a translation into the matrix as the
third column, and we want our matrices square so we an invert them. Our
implementation computes the product of a cMatrix3 and cVector3 by doing the 
standard multiplication of a 4x4 matrix times a 1x4 column vector.  */

//Preliminary declarations involving types we'll specify later on in this file.========
class cMatrix3; //So cVector3 can mention operator*(cMatrix3, cVector3)
class cSpin; //A container to hold a magnitude and an axis for 3D roations.
//Typedef of the generic vector and matrix classes we'll use, so we don't really need
//the "3" at the end.
typedef class cVector3 cVector;
typedef class cMatrix3 cMatrix;
//End preliminary type declarations.=================================================


class cVector3
{
public:
	static const Real PRACTICALLY_ZERO; /* Used by the isZero method to treat as zero
		vectors the occasional neglibible rounded off remains of a vector that should be 
		zero. See isPracticallyZero below. */
	static const Real PRACTICALLY_PARALLEL_COSINE; /* Used when you dot product
		two unit vectors together; if they're "practiclaly parallel" the dot product is
		greater than this. */
	static const cVector3 ZEROVECTOR;
	static const cVector3 XAXIS;
	static const cVector3 YAXIS;
	static const cVector3 ZAXIS;

private:
	Real _x, _y, _z;
	friend cMatrix3;
public:
	cVector3(){_x = _y = _z = 0.0;} //The 0 vector. Use  default copy constructor and =. 
	cVector3(Real ix, Real iy, Real iz):_x(ix),_y(iy),_z(iz){}
	cVector3(Real ix, Real iy):_x(ix),_y(iy),_z(0.0){}
	cVector3(const cVector3 &v){_x = v._x; _y = v._y; _z = v._z;}
//Mutators
	void set(Real ix, Real iy, Real iz){_x = ix; _y = iy; _z = iz;}
	void set(Real ix, Real iy){_x = ix; _y = iy;}//leave z alone
	void setZ(Real iz){_z = iz;}
	void setZero(){_x = 0.0; _y = 0.0; _z = 0.0;}
	cVector3 normalize(); //Make yourself of unit length and return self.
	Real normalizeAndReturnMagnitude(); //Make yourself of unit length and return mag.
	cVector3 roundOff(); /* Get rid of any component with size less than SMALL_REAL.  This
		can be useful if you have some nagging little roundoff residue in a compoment that
		should be 0. Return self. */
	void setMagnitude(Real mag); //Make your length be mag
	void apply(const cMatrix3 &M); //Apply M to self.
	cVector3 turn(Real angle); //Means rotate around z axis.
	cVector3 rotate(const cSpin &spin);  //Can be around any axis.
//Accessors 
	Real x()const {return _x;}
	Real y()const {return _y;}
	Real z()const {return _z;}
	Real magnitude() const; //Return the cVector3's length.
	Real distanceTo(const cVector3 &vect) const;
//	Real angle(cVector3 standarddir=cVector3(1.0, 0.0, 0.0)) const 
//		{return angleBetween(*this, standarddir);} /* Returns an angle between 0 and PI.*/
	Real angle()const; //Collapse to 2D and get that angle.
	cVector3 direction()const; //Return a unit vector in your direction.
	bool isZero()const{return _x == 0.0 && _y == 0.0 && _z == 0.0;}
	bool isPracticallyZero()const{return magnitude() < PRACTICALLY_ZERO;} 
	bool isPracticallyEqualTo(const cVector3 &u)const
		{return (*this-u).magnitude() < PRACTICALLY_ZERO;} 
	/* Sometimes if I do something like u = cVector3(1,0) - cVector3(1,0),
			u I can end up with a nonzero magnitude.  So I use this to winnow out
			inaccurate vectors that should be zero. */
//Other Methods
	friend Real distance(const cVector3 &u, const cVector3 &v); //Distance
	friend Real angleBetween(const cVector3 &u, const cVector3 &v); /* The angle of u
		minus the angle of v, that is, the counterclockwise	angle gotten by
		turning from v to u.  Value between 0 and 2*PI. */
	cVector3 defaultNormal(); //Useful sometimes to arbitrarily pick a normal
//Overloaded Operators
	cVector3 operator=(const cVector3 &u){_x=u._x; _y=u._y; _z=u._z; return *this;}
	cVector3 operator+=(const cVector3 &u){_x+=u._x; _y+=u._y; _z+=u._z; return *this;}
	cVector3 operator-=(const cVector3 &u){_x-=u._x; _y-=u._y; _z-=u._z; return *this;}
	cVector3 operator*=(Real f){_x*=f; _y*=f;  _z*=f; return *this;}
	cVector3 operator/=(Real f);
//Factory Method
	static cVector3 randomUnitVector();
//Overloaded Friend operators
	friend cVector3 operator*(const cVector3 &u, Real f); //Scalar prod
	friend cVector3 operator*(Real f, const cVector3 &u); //Scalar prod
	friend cVector3 operator/(const cVector3 &u, Real f); //Scalar division
	friend cVector3 operator+(const cVector3 &u, const cVector3 &v); //cVector3 sum
	friend cVector3 operator-(const cVector3 &u, const cVector3 &v);//cVector3 difference.
	friend cVector3 operator-(const cVector3 &u){return cVector3(-u._x,-u._y,-u._z);}
		//Unary -
	friend Real operator%(const cVector3 &u, const cVector3 &v); //Dot product
	friend cVector3 operator*(const cVector3 &u, const cVector3 &v); //cross product
	friend int operator==(const cVector3 &u, const cVector3 &v) //equality predicate.
		{return (u._x == v._x && u._y == v._y && u._z == v._z);}
	friend int operator!=(const cVector3 &u, const cVector3 &v) //inequality predicate.
		{return !(u._x == v._x && u._y == v._y && u._z == v._z);}
	friend cVector3 operator*(const cMatrix3 &M, const cVector3 &V);
//Serialization
	friend ostream& operator<<(ostream& ar, const cVector3 &v);
	friend istream& operator>>(istream& ar, cVector3 & v);
//OpenGL Methods
	void glVertex(){::glVertex3d((GLdouble)_x, (GLdouble)_y, (GLdouble)_z);}
};

//Helper class
class cSpin
{
public:
private:
	Real _spinangle;
	cVector3 _spinaxis;
public:
	cSpin():_spinangle(0.0), _spinaxis(cVector3(0.0, 0.0, 1.0)){}
	cSpin(cVector spinvector);
	cSpin(Real spinangle, cVector spinaxis);
	cSpin(Real angle):_spinangle(angle),_spinaxis(cVector3(0.0, 0.0, 1.0)){}
	cSpin(const cSpin &spin){*this = spin;}
	Real spinangle()const{return _spinangle;}
	cVector3 spinaxis()const{return _spinaxis;}
	void setZero(){_spinangle = 0.0; _spinaxis = cVector3(0.0, 0.0, 1.0);}
	cSpin& operator=(const cSpin &spin){_spinangle = spin._spinangle;
		_spinaxis = spin._spinaxis;	return *this;}
	friend cSpin operator*(Real f, const cSpin &spin){cSpin fspin(spin);
		fspin._spinangle*=f; return fspin;}
	cSpin operator*=(const cMatrix3 &m){_spinaxis = m *_spinaxis; return *this;}
	friend ostream& operator<<(ostream& ar, const cSpin &v);
	friend istream& operator>>(istream& ar, cSpin & v);
};		

// ----- Three Dimensional Matrixes, or Transformations -------

class cMatrix3
{
private:
	Real _mat[4][4];
		/*Note that _mat[i][j] means the entry in the i_th row and the j_th
		column, so this is similar to usual matrix notation like m_ij.  One
		minor confusion is that C language indices start at 0, but matrix
		indices usually start at 1.*/
	friend cVector3;
public:
//Constructor and copy
	cMatrix3(){identity();} //Call the Identity operation at construction.
	cMatrix3(Real e00, Real e10, Real e20, Real e30, Real e01, Real e11, Real e21, Real e31,
		Real e02, Real e12, Real e22, Real e32, Real e03, Real e13, Real e23, Real e33); /* We
		feed the arguments in column order, that is the first four are the first column, and so on. */
	cMatrix3(Real *vert); //Assume vert is an array of 16.
	cMatrix3(cVector3 tangent, cVector3 normal, cVector3 binormal,
		 cVector3 position); /*This will set the cols to tangent, normal and position. */
	cMatrix3(const cMatrix3 &M);
	void copy(const cMatrix3 &M);
	cMatrix3& operator=(const cMatrix3 &M);
//Accessor
	Real* elements(){return (Real*)_mat;} /* This returns the elements in row order, that is, the
		top row, followed by the next row and so on.  OpenGL uses elements in column order, so
		use a call like loadMatrix(M.transpose().elements()) to load into OpenGL.*/
	cVector3 column(int i)const;
	cVector3 lastColumn()const{return column(3);}
	Real scalefactor()const; //How much scaling does this do?  Assume isotropic, so test on unit vector of XAXIS

//Mutators
	void identity(); //Set to the identity transformation.
		/* The following three friend operations create special kinds
			of matrices*/
	void set(Real e00, Real e10, Real e20, Real e30, Real e01, Real e11, Real e21, Real e31,
		Real e02, Real e12, Real e22, Real e32, Real e03, Real e13, Real e23, Real e33); /* We
		feed the arguments in column order, that is the first four are the first column, and so on. */
	void set(Real *vert); //Assume vert is an array of 16.
	void setColumn(int j, const cVector3 &u); //j is 0, 1, 2, or 3.
	void setLastColumn(const cVector3 &u){setColumn(3,u);}
	void setZTranslation(Real zpos){setLastColumn(cVector3(lastColumn().x(), lastColumn().y(), zpos));}
	void translate(const cVector3 &trans);
	void orthonormalize(); //Make first three columns an orthonormal basis.
//Matrix methods
	cMatrix3 transpose();
	cMatrix3 inverse();/*This only works if we assume the matrix is a rigid-body matrix, 
		i.e. that its cVector3 columns are orthonormal unit vectors as are its cVector3
		rows (,ignoring the foruth column.)  You can have any translation you like in the fourth column. */
	cMatrix3 normalTransformation();
//Overloaded operators
	cMatrix3& operator*=(const cMatrix3 &M);
	friend cMatrix3 operator*(const cMatrix3 &M,
		const cMatrix3 &N);
	friend cVector3 operator*(const cMatrix3 &M, const cVector3 &V);
	friend cMatrix3 operator*(Real f, cMatrix3 M){for (int i=0; i<16; i++)M.elements()[i]*=f; return M;}
//Factory methods
	/* The following  operations create special matrices*/
	static cMatrix3 identityMatrix(){return cMatrix3();}
	static cMatrix3 xRotation(Real ang);
	static cMatrix3 yRotation(Real ang);
	static cMatrix3 zRotation(Real ang);
	static cMatrix3 rotation(Real ang){return rotation(cSpin(ang));}
	static cMatrix3 rotation(const cSpin &spin); 
	static cMatrix3 rotationFromUnitToUnit(const cVector3 &u, const cVector3 &v); /* Assume that
		u and v are unit vectors.  This gives a matrix that rotates u into v. */
	static cMatrix3 translation(const cVector3 &trans);
	static cMatrix3 scale(Real size);
	static cMatrix3 scale(Real xscale, Real yscale, Real zscale);
//Serialization
	friend ostream& operator<<(ostream& ar, const cMatrix3 & M);
	friend istream& operator>>(istream& ar, cMatrix3 &M);
};

class cLine
{
public:
	cVector _origin;
	cVector _tangent;
	cLine():_tangent(1,0,0){} //Default origin is (0,0,0)
	cLine(const cVector &origin, const cVector &tangent):_origin(origin),_tangent(tangent){}
	cLine(const cLine &line):_origin(line._origin),_tangent(line._tangent){}
	Real distanceTo(const cVector &point)const;
	Real lineCoord(const cVector &point)const; /* Project point onto the line, then
		take it's coord wiht the line numbered so _origin is 0.0, _origin+_tangent is 1.0, 
		and so on. */
};

class cPlane
{
public:
	cVector _origin;
	cVector _binormal;
	cPlane():_binormal(0,0,1){} //Default origin is (0,0,0)
	cPlane(const cVector &origin, const cVector &binormal):_origin(origin),_binormal(binormal){}
	cPlane(const cPlane &plane):_origin(plane._origin),_binormal(plane._binormal){}
	cVector project(const cVector &point)const;
	cVector intersect(const cLine &line)const;
};


#endif //VECTOR_HPP
